# -*- coding: utf-8 -*- 
# Time: 2021-06-29 17:45
# Copyright (c) 2021
# author: Euraxluo


from datetime import datetime
import typing
from fastapi import APIRouter, Query, Body, Path

import apscheduler.util
from apscheduler.triggers.cron import CronTrigger
from apscheduler.triggers.date import DateTrigger
from apscheduler.triggers import base as BaseTrigger
from apscheduler.triggers.interval import IntervalTrigger

from fast_job.job_schedule import schedule, JobSchedule, fast_job_api_router
from fast_job.schema import Response, SchedulesBase, TaskWorkRecord, Job, ExceptionCode, SuccessMessage

job_api_prefix = "/schedule"


class SecondsCronTrigger(CronTrigger):
    @classmethod
    def seconds_crontab(cls, expr, timezone=None):
        values = expr.split()
        if len(values) != 7:
            raise ValueError("Wrong number of fields; got {}, expected 7".format(len(values)))

        return cls(second=values[0], minute=values[1], hour=values[2], day=values[3], month=values[4],
                   day_of_week=values[5], year=values[6], timezone=timezone)


# interval Fixed interval scheduling
@fast_job_api_router.post(job_api_prefix + "/interval", response_model=Response, summary="Interval a periodic schedule that runs in a cycle")
async def interval(
        job_id: str = Body(..., description="job_id"),
        task_id: str = Body(..., description="task_id"),
        run_time: typing.Union[str, datetime] = Body(datetime.now(), description="First run time"),
        seconds: int = Body(120, description="Cycle interval / second, default 120S"),
        tag: str = 'default'):
    schedule.logger.info({"Add Interval Scheduling mode Task": {"job_id": job_id, "task_id": task_id, "run_time": run_time, "seconds": seconds, "tag": tag}})
    job = schedule.get_job(job_id=schedule.job_key(job_id=job_id, tag=tag))
    if job:
        return Response(code=ExceptionCode.ScheduleCreateExistsFailed.value, message=f"{job_id} job already exists")

    if run_time == '' or (isinstance(run_time, datetime) and run_time < datetime.now()):
        run_time = apscheduler.util.undefined

    job = schedule.add_job(JobSchedule().task_scheduler,
                           'interval',
                           args=(job_id, task_id, tag),
                           seconds=seconds,
                           id=schedule.job_key(job_id=job_id, tag=tag),
                           next_run_time=run_time,
                           max_instances=1000,
                           misfire_grace_time=3600
                           )
    return Response(message=SuccessMessage.JobSetIntervalSuccess.value, data={"job": Job(job)})


# date Run only once at a particular point in time
@fast_job_api_router.post(job_api_prefix + "/datetime", response_model=Response, summary="Scheduled scheduling that runs at a specified point in time")
async def datetimer(
        job_id: str = Body(..., description="job_id"),
        task_id: str = Body(..., description="task_id"),
        run_time: typing.Union[str, datetime] = Body(datetime.now(), description="Run time specified"),
        tag: str = 'default'):
    schedule.logger.info({"Add datetime scheduling mode Task": {"job_id": job_id, "task_id": task_id, "run_time": run_time, "tag": tag}})
    job = schedule.get_job(job_id=schedule.job_key(job_id=job_id, tag=tag))
    if job:
        return Response(code=ExceptionCode.ScheduleCreateExistsFailed.value, message=f"{job_id} job already exists")

    if run_time == '' or (isinstance(run_time, datetime) and run_time < datetime.now()):
        run_time = datetime.now()

    job = schedule.add_job(JobSchedule().task_scheduler,
                           'date',
                           args=(job_id, task_id, tag),
                           run_date=run_time,
                           id=schedule.job_key(job_id=job_id, tag=tag),  # job ID
                           max_instances=1000,
                           misfire_grace_time=3600
                           )
    return Response(message=SuccessMessage.JobSetDatetimeSuccess.value, data={"job": Job(job)})


# cron More flexible scheduled tasks can use the crontab expression
@fast_job_api_router.post(job_api_prefix + "/cron", response_model=Response, summary="Scheduled scheduling using crontab expression")
async def cron(
        job_id: str = Body(..., description="job_id"),
        task_id: str = Body(..., description="task_id"),
        run_time: typing.Union[str, datetime] = Body(datetime.now(), description="First run time"),
        crontab: str = Body('*/10 * 8-20 * * * *', description="crontab expression"),
        tag: str = 'default'):
    schedule.logger.info({"Add cron scheduled Task": {"job_id": job_id, "task_id": task_id, "run_time": run_time, "crontab": crontab, "tag": tag}})
    job = schedule.get_job(job_id=schedule.job_key(job_id=job_id, tag=tag))
    if job:
        return Response(code=ExceptionCode.ScheduleCreateExistsFailed.value, message=f"{job_id} job already exists")

    if run_time == '' or (isinstance(run_time, datetime) and run_time < datetime.now()):
        run_time = apscheduler.util.undefined

    job = schedule.add_job(JobSchedule().task_scheduler,
                           SecondsCronTrigger.seconds_crontab(crontab),
                           args=(job_id, task_id, tag),
                           id=schedule.job_key(job_id=job_id, tag=tag),
                           next_run_time=run_time,
                           max_instances=1000,
                           misfire_grace_time=3600
                           )
    return Response(message=SuccessMessage.JobSetCronSuccess.value, data={"job": Job(job)})


@fast_job_api_router.get(job_api_prefix + "/scheduling", response_model=Response, summary="Get information about all scheduling jobs")
async def scheduling(tag: str = 'default'):
    all_schedule: typing.List[SchedulesBase] = []
    for job in schedule.get_jobs():
        if schedule.check_job_key(job_id=job.id, tag=tag):
            all_schedule.append(SchedulesBase(tag=tag, job_id=schedule.get_real_job_id(job.id), func_name=job.func_ref, func_args=job.args, cron_model=str(job.trigger), next_run_time=job.next_run_time))
    schedule.logger.info({"Get information about all scheduling jobs": {"tag": tag, "all_job": all_schedule}})
    if len(all_schedule) == 0:
        return Response(code=ExceptionCode.ScheduleNotFoundFailed.value, message=f"job not found")
    return Response(message=SuccessMessage.JobGetAllSuccess.value, data={"schedules": all_schedule})


@fast_job_api_router.get(job_api_prefix + "/job", response_model=Response, summary="Get information about the specified job")
async def job_info(
        job_id: str = Query(None, description="任务id"),
        tag: str = 'default'):
    schedule.logger.info({"Get information about the specified job": {"tag": tag, "job_id": job_id}})
    job = schedule.get_job(job_id=schedule.job_key(job_id=job_id, tag=tag))
    if job:
        return Response(message=SuccessMessage.JobGetSuccess.value, data=SchedulesBase(tag=tag, job_id=job.id, func_name=job.func_ref, func_args=job.args, cron_model=str(job.trigger), next_run_time=job.next_run_time).dict())
    return Response(code=ExceptionCode.ScheduleNotFoundFailed.value, message=f"not found job {job_id},tag {tag},job_key {schedule.job_key(job_id=job_id, tag=tag)}")


@fast_job_api_router.post(job_api_prefix + "/remove/{job_id}", response_model=Response, summary="Remove a task based on the job ID")
async def remove(
        job_id: str = Path(..., description="job_id"),
        tag: str = 'default'):
    schedule.logger.info({"Remove a task based on the job ID": {"tag": tag, "job_id": job_id}})
    remove_work_rdb_resp = schedule.remove_work_rdb(job_id=job_id, tag=tag)
    job = schedule.get_job(job_id=schedule.job_key(job_id=job_id, tag=tag))
    if not job:
        return Response(code=ExceptionCode.ScheduleNotFoundFailed.value, message=f"not found job {job_id},work_log {remove_work_rdb_resp}")

    schedule.remove_job(schedule.job_key(job_id=job_id, tag=tag))
    return Response(message=SuccessMessage.JobRemoveSuccess.value, data={"job": Job(job), "work_log": remove_work_rdb_resp})


@fast_job_api_router.post(job_api_prefix + "/pause/{job_id}", response_model=Response, summary="Pause a job by job ID")
async def pause(
        job_id: str = Path(..., description="job_id"),
        tag: str = 'default'):
    schedule.logger.info({"Pause a job by job ID": {"tag": tag, "job_id": job_id}})
    job = schedule.get_job(job_id=schedule.job_key(job_id=job_id, tag=tag))
    if not job:
        return Response(code=ExceptionCode.ScheduleNotFoundFailed.value, message=f"not found job {job_id}")
    schedule.pause_job(schedule.job_key(job_id=job_id, tag=tag))
    return Response(message=SuccessMessage.JobPauseSuccess.value, data={"job": Job(job)})


@fast_job_api_router.post(job_api_prefix + "/resume/{job_id}", response_model=Response, summary="Resume a job by job ID")
async def resume(
        job_id: str = Path(..., description="job_id"),
        tag: str = 'default'):
    schedule.logger.info({"Resume a job by job ID": {"tag": tag, "job_id": job_id}})
    job = schedule.get_job(job_id=schedule.job_key(job_id=job_id, tag=tag))
    if not job:
        return Response(code=ExceptionCode.ScheduleNotFoundFailed.value, message=f"not found job {job_id}")
    schedule.resume_job(schedule.job_key(job_id=job_id, tag=tag))
    return Response(message=SuccessMessage.JobResumeSuccess.value, data={"job": Job(job)})


@fast_job_api_router.post(job_api_prefix + "/modify", response_model=Response, summary="Modify a task based on the job ID")
async def modify(
        job_id: str = Body("", description="job_id"),
        task_id: typing.Optional[str] = Body("", description="task_id"),
        trigger: typing.Optional[str] = Body("interval", description="(e.g. `date`, `interval` or `cron`)"),
        run_time: typing.Union[str, datetime] = Body(datetime.now(), description="Expected first run time"),
        crontab: str = Body('*/10 * 8-20 * * * *', description="crontab expression"),
        seconds: int = Body(120, description="Cycle mode interval, unit / second, default 120S"),
        tag: str = 'default'):
    schedule.logger.info({"Modify a task based on the job ID": {"job_id": job_id, "task_id": task_id, "trigger": trigger, "run_time": run_time, "crontab": crontab, "seconds": seconds, "tag": tag}})
    job = schedule.get_job(job_id=schedule.job_key(job_id=job_id, tag=tag))
    if not job:
        return Response(code=ExceptionCode.ScheduleNotFoundFailed.value, message=f"not found job {job_id}")

    if run_time == '' or (isinstance(run_time, datetime) and run_time < datetime.now()):
        run_time = apscheduler.util.undefined

    new_trigger: BaseTrigger = None
    if trigger == "date":
        new_trigger = DateTrigger(run_date=run_time)
    elif trigger == "interval":
        new_trigger = IntervalTrigger(seconds=seconds)
    elif trigger == "cron":
        new_trigger = SecondsCronTrigger.seconds_crontab(crontab)
    elif trigger is not None:
        return Response(code=ExceptionCode.ScheduleNotFoundFailed.value, message=f"modify failed,because trigger wrong")
    args = job.args
    if task_id:
        args = (args[0], job_id, task_id, tag)
    try:
        job = schedule.modify_job(job_id=schedule.job_key(job_id=job_id, tag=tag), trigger=new_trigger, next_run_time=run_time, args=args)
        schedule.resume_job(schedule.job_key(job_id=job_id, tag=tag))  # might delete it failed
    except Exception as e:
        return Response(message=f"Task modification failed:{e}")
    return Response(message=SuccessMessage.JobModifySuccess.value, data={"job": Job(job)})


@fast_job_api_router.get(job_api_prefix + "/log", response_model=Response, summary="Get execution records corresponding to the job ID")
async def log(
        job_id: str = Query(..., description="job_id"),
        start: int = Query(0, description="Start index, starting at 0, default 0"),
        end: int = Query(99, description="End index, default is 99"),
        tag: str = 'default'):
    schedule.logger.info({"Get execution records corresponding to the job ID": {"job_id": job_id, "start": start, "end": end, "tag": tag}})
    job = schedule.get_job(job_id=schedule.job_key(job_id=job_id, tag=tag))
    if not job:
        return Response(code=ExceptionCode.ScheduleNotFoundFailed.value, message=f"not found job {job_id}")

    work_records = schedule.list_work_record(job_id=job_id, tag=tag, start=start, end=end)
    return Response(data={"log": work_records})


@fast_job_api_router.post(job_api_prefix + "/retry", response_model=Response, summary="Retry execution records based on the job ID and other parameters")
async def retry(work: TaskWorkRecord = Body(..., description="Execute record structure")):
    schedule.logger.info({"Retry execution records based on the job ID and other parameters": {"work": work.dict()}})
    try:
        result = schedule.task_retry(work=work)
    except Exception as e:
        return Response(code=ExceptionCode.ScheduleRetryFailed.value, message=str(e))
    return Response(message=SuccessMessage.JobRetrySuccess.value, data={"result": result})


@fast_job_api_router.get(job_api_prefix + "/tasks", response_model=Response, summary="all load tasks")
async def tasks():
    return Response(data=schedule.__task_manage__)



import os
import uuid
import ipaddress
from socket import gethostbyname, gethostname
from .about import NAME, VERSION, __version__
from .sonyflake import SonyFlake




def lower_16bit_ip_mac_pid() -> int:
    """
    :return:
    """
    ip: ipaddress.IPv4Address = ipaddress.ip_address(gethostbyname(gethostname()))
    ip_bytes = ip.packed
    res = hash(str([(ip_bytes[2] << 8) + ip_bytes[3], uuid.getnode(), os.getpid()])) & 0xffff
    return res


sony_flake = SonyFlake(machine_id=lower_16bit_ip_mac_pid)


def SonyFlakeId():
    return sony_flake.next_id()



__all__ = ["SonyFlakeId"]



# -*- coding: utf-8 -*- 
# Time: 2022-02-15 10:56
# Copyright (c) 2022
# author: Euraxluo

import datetime
import ipaddress
from socket import gethostbyname, gethostname
from threading import Lock
from time import sleep
from typing import Callable, Dict, Optional

BIT_LEN_TIME = 39
BIT_LEN_SEQUENCE = 8
BIT_LEN_MACHINE_ID = 63 - (BIT_LEN_TIME + BIT_LEN_SEQUENCE)
UTC = datetime.timezone.utc
SONYFLAKE_EPOCH = datetime.datetime(2014, 9, 1, 0, 0, 0, tzinfo=UTC)


def lower_16bit_private_ip() -> int:
    """
    Returns the lower 16 bits of the private IP address.
    """
    ip: ipaddress.IPv4Address = ipaddress.ip_address(gethostbyname(gethostname()))
    ip_bytes = ip.packed
    return (ip_bytes[2] << 8) + ip_bytes[3]


class SonyFlake:
    """
    The distributed unique ID generator.
    """

    _start_time: int
    _machine_id: int

    def __new__(
        cls,
        start_time: Optional[datetime.datetime] = None,
        machine_id: Optional[Callable[[], int]] = None,
        check_machine_id: Optional[Callable[[int], bool]] = None,
    ):
        if start_time and datetime.datetime.now(UTC) < start_time:
            return None
        instance = super().__new__(cls)
        if machine_id is not None:
            instance._machine_id = machine_id()
        else:
            instance._machine_id = lower_16bit_private_ip()
        if check_machine_id is not None:
            if not check_machine_id(instance._machine_id):
                return None
        return instance

    def __init__(
        self,
        start_time: Optional[datetime.datetime] = None,
        machine_id: Optional[Callable[[], int]] = None,
        check_machine_id: Optional[Callable[[int], bool]] = None,
    ) -> None:
        """
        Create a new instance of `SonyFlake` unique ID generator.

        `start_time` is the time since which the SonyFlake time is
        defined as the elapsed time.

        * If `start_time` is 0, the start time of the SonyFlake is set
        to _"`2014-09-01 00:00:00 +0000 UTC`"_.

        * If `start_time` is ahead of the current time, SonyFlake is
        not created.

        `machine_id` returns the unique ID of the SonyFlake instance.

        * If `machine_id` returns an error, SonyFlake is not created.

        * If `machine_id` is nil, default `machine_id` is used.

        * Default `machine_id` returns the lower 16 bits of the
        private IP address.

        `check_machine_id` validates the uniqueness of the machine ID.

        * If `check_machine_id` returns `False`, SonyFlake is not
        created.

        * If `check_machine_id` is `None`, no validation is done.
        """
        if start_time is None:
            start_time = SONYFLAKE_EPOCH
        self.mutex = Lock()
        self._start_time = self.to_sonyflake_time(start_time)
        self.elapsed_time = self.current_elapsed_time()
        self.sequence = (1 << BIT_LEN_SEQUENCE) - 1
        if not hasattr(self, "_machine_id"):
            self._machine_id = machine_id and machine_id() or lower_16bit_private_ip()

    @staticmethod
    def to_sonyflake_time(given_time: datetime.datetime) -> int:
        """
        Convert a `datetime.datetime` object to SonyFlake's time
        value.
        """
        return int(given_time.timestamp() * 100)

    @property
    def start_time(self) -> int:
        return self._start_time

    @property
    def machine_id(self) -> int:
        return self._machine_id

    def current_time(self) -> int:
        """
        Get current UTC time in the SonyFlake's time value.
        """
        return self.to_sonyflake_time(datetime.datetime.now(UTC))

    def current_elapsed_time(self) -> int:
        """
        Get time elapsed since the SonyFlake ID generator was
        initialised.
        """
        return self.current_time() - self.start_time

    def next_id(self) -> int:
        """
        Generates and returns the next unique ID.

        Raises a `TimeoutError` after the `SonyFlake` time overflows.
        """
        mask_sequence = (1 << BIT_LEN_SEQUENCE) - 1
        with self.mutex:
            current_time = self.current_elapsed_time()
            if self.elapsed_time < current_time:
                self.elapsed_time = current_time
                self.sequence = 0
            else:
                self.sequence = (self.sequence + 1) & mask_sequence
                if self.sequence == 0:
                    self.elapsed_time += 1
                    overtime = self.elapsed_time - current_time
                    sleep(self.sleep_time(overtime))
            return self.to_id()

    def to_id(self) -> int:
        if self.elapsed_time >= (1 << BIT_LEN_TIME):
            raise TimeoutError("Over the time limit!")
        time = self.elapsed_time << (BIT_LEN_SEQUENCE + BIT_LEN_MACHINE_ID)
        sequence = self.sequence << BIT_LEN_MACHINE_ID
        return time | sequence | self.machine_id

    @staticmethod
    def sleep_time(duration: int) -> float:
        """
        Calculate the time remaining until generation of new ID.
        """
        return (
            duration * 10 - (datetime.datetime.now(UTC).timestamp() * 100) % 1
        ) / 100

    @staticmethod
    def decompose(_id: int) -> Dict[str, int]:
        """
        Decompose returns a set of SonyFlake ID parts.
        """
        mask_sequence = ((1 << BIT_LEN_SEQUENCE) - 1) << BIT_LEN_MACHINE_ID
        mask_machine_id = (1 << BIT_LEN_MACHINE_ID) - 1
        msb = _id >> 63
        time = _id >> (BIT_LEN_SEQUENCE + BIT_LEN_MACHINE_ID)
        sequence = (_id & mask_sequence) >> BIT_LEN_MACHINE_ID
        machine_id = _id & mask_machine_id
        return {
            "id": _id,
            "msb": msb,
            "time": time,
            "sequence": sequence,
            "machine_id": machine_id,
        }



# -*- coding: utf-8 -*- 
# Time: 2022-02-18 19:38
# Copyright (c) 2022
# author: Euraxluo

from fast_job.job_api import schedule, fast_job_api_router

__version__ = "0.1.11"
__description__ = """Provides scheduling apis and scheduling and task-related services"""
__all__ = ["fast_job_api_router", "schedule"]



# -*- coding: utf-8 -*- 
# Time: 2021-06-29 17:56
# Copyright (c) 2021
# author: Euraxluo

import datetime
from typing import List, Any, Mapping, Optional, Tuple
from pydantic import BaseModel
from apscheduler.job import Job as scheduleJob

from enum import Enum


class ExceptionCode(Enum):
    ScheduleFailed = 8000
    ScheduleCreateExistsFailed = 8100
    ScheduleNotFoundFailed = 8101
    ScheduleRetryFailed = 8102


class SuccessMessage(Enum):
    JobSetCronSuccess = "Cron scheduler creation success"
    JobSetDatetimeSuccess = "Datetime scheduler creation success"
    JobSetIntervalSuccess = "Interval scheduler creation success"

    JobGetAllSuccess = "Get all task information is get success"
    JobGetSuccess = "Get task information is get success"

    JobModifySuccess = "Modify the task scheduling information success"
    JobResumeSuccess = "Resume the task scheduling information success"
    JobPauseSuccess = "Pause the task scheduling information success"
    JobRemoveSuccess = "Remove the task scheduling information success"

    JobRetrySuccess = "Task Retry succeeded"


class SchedulesBase(BaseModel):
    tag: str
    job_id: str
    func_name: str
    func_args: Tuple
    cron_model: str
    next_run_time: Optional[datetime.datetime]


class TaskWorkRecord(BaseModel):
    run_id: str
    job_id: str
    task_id: str
    tag: str
    success: bool = False
    result: Any = None
    error: Any = None
    task_args: List[Any]
    task_kwargs: Mapping[str, Any]
    start_time: float
    end_time: float
    process_time: float


class Response(BaseModel):
    """
    API Response
    """
    code: Optional[int] = 200
    message: Optional[str] = ""
    data: Optional[dict] = {}


class Job(object):
    def __init__(self, job: scheduleJob):
        from fast_job.job_schedule import schedule
        tag, job_id = schedule.get_job_tag_with_job_id(job.id)
        self.tag = tag
        self.id = job_id
        self.trigger = str(job.trigger)
        self.func_ref = job.func_ref
        self.args = job.args
        self.kwargs = job.kwargs
        self.name = job.name
        self.misfire_grace_time = job.misfire_grace_time
        self.next_run_time = job.next_run_time



# -*- coding: utf-8 -*- 
# Time: 2022-02-15 10:57
# Copyright (c) 2022
# author: Euraxluo

import redis
import time
import math
import threading
import typing
from redis import Redis


class _WatchThread(threading.Thread):
    def __init__(self, target, args=(), kwargs=None):
        super(_WatchThread, self).__init__()
        if kwargs is None:
            kwargs = {}
        self.func = target
        self.args = args
        self.kwargs = kwargs
        self.result = None

    def run(self):
        # 接受返回值
        self.result = self.func(*self.args, **self.kwargs)

    def get_result(self, default=None, transform=lambda x: x):
        # 线程不结束,返回值为None
        # noinspection PyBroadException
        try:
            return transform(self.result)
        except Exception as _:
            return transform(default)


# 加载脚本并执行的函数
def _script_load(script):
    sha = [None]

    def call(conn: Redis, keys=None, args=None, force_eval=False):
        if args is None:
            args = []
        if keys is None:
            keys = []
        if not force_eval:
            # 加载并缓存校验和
            if not sha[0]:
                sha[0] = conn.execute_command("SCRIPT", "LOAD", script, parse="LOAD")
                if isinstance(conn, redis.client.Pipeline):
                    sha[0] = conn.execute()[0]
            try:
                return conn.execute_command("EVALSHA", sha[0], len(keys), *(keys + args))
            except redis.exceptions.ResponseError as msg:
                if not msg.args[0].startswith("NOSCRIPT"):
                    raise
        # 如果需要强制执行或者脚本接收到错误时，会使用eval执行脚本，eval执行完脚本后，会把脚本的sha1值缓存下来
        return conn.execute_command("EVAL", script, len(keys), *(keys + args))

    return call


"""添加互斥锁"""
_acquire_lock_with_timeout_lua = _script_load(
    """
    if (redis.call('exists',KEYS[1]) == 0) then
        return redis.call('setex',KEYS[1],unpack(ARGV))
    end
    """
)


def acquire_lock_with_timeout(conn: Redis, lock_name: str, identifier: str, acquire_timeout: int = 10, lock_timeout: int = 10):
    lock_name = lock_name
    lock_timeout = int(math.ceil(lock_timeout))

    acquired = False
    end = time.time() + acquire_timeout
    while time.time() < end and not acquired:
        acquired = _acquire_lock_with_timeout_lua(conn, [lock_name], [lock_timeout, identifier]) == 'OK'
        time.sleep(0.001 * (not acquired))
    return acquired and identifier


"""释放互斥锁"""
_release_lock_lua = _script_load(
    """
    if (redis.call('get',KEYS[1])== ARGV[1]) then
        return redis.call('del',KEYS[1]) or true
    end
    """
)


def release_lock(conn: Redis, lock_name: str, identifier: str):
    lock_name = lock_name
    return _release_lock_lua(conn, [lock_name], [identifier])


"""为可重入锁续约"""
_renew_re_entrant_lock_lua = _script_load(
    # KEYS : lock_name
    # ARGV : lock_timeout,id
    """
    if (redis.call('hexists',KEYS[1],ARGV[2])==1) then
        redis.call('pexpire',KEYS[1],ARGV[1])
        return 1
    end
    return 0
    """
)


def _renew_re_entrant_lock(conn: Redis, lock_name: str, identifier: str, lock_timeout: int):
    """
    为锁续约
    :param conn:redis链接
    :param lock_name:lock_key
    :param identifier:lock_唯一id
    :param lock_timeout: 上锁超时时间ms
    :return:
    """
    while _renew_re_entrant_lock_lua(conn, [lock_name], [lock_timeout, identifier]):
        # 若未续约成功则直接退出
        time.sleep(lock_timeout / 3000)


"""添加可重入锁"""


def _watch_operation_time(limit_time: int):
    """
    监视运行时间
    :param limit_time:
    :return:
    """

    def functions(func):
        def run(*params):
            watch_thread = _WatchThread(target=func, args=params)
            watch_thread.setDaemon(True)
            watch_thread.start()
            for i in range(int(limit_time // 0.001)):
                if watch_thread.get_result(default=-1) != -1:
                    return watch_thread.get_result
                time.sleep(0.001)
            time.sleep(round(limit_time - int(limit_time // 0.001) * 0.001, 4))
            return watch_thread.get_result  # 时间结束后,返回thread对象

        return run

    return functions


_acquire_re_entrant_lock_with_timeout_lua = _script_load(
    # KEYS : lock_name
    # ARGV : lock_timeout,id

    # 如果exists 结果为 0 ，标识无锁，加锁，将本线程获取的id加到hset中，重置过期时间，返回nil
    # 相同线程加锁：exists 为 1，此时判断本进程id 是否存在于hset中，若存在，将信号量+1，并重置过期时间，返回nil
    # 其他线程加锁：直接pttl key，返回剩余过期时间，脚本未返回nil，加锁失败
    """
    if (redis.call('exists',KEYS[1]) == 0) then
        redis.call('hset',KEYS[1],ARGV[2],1)
        redis.call('pexpire',KEYS[1],ARGV[1])
        return nil
    end
    if (redis.call('hexists',KEYS[1],ARGV[2])==1) then
        redis.call('hincrby',KEYS[1],ARGV[2],1)
        redis.call('pexpire',KEYS[1],ARGV[1])
        return nil
    end
    return redis.call('pttl',KEYS[1])
    """
)


def acquire_re_entrant_lock_with_timeout(conn: Redis, lock_name: str, identifier: str, acquire_timeout: int = -1, lock_timeout: int = -1) -> typing.Callable:
    """
    :param conn: Redis链接
    :param lock_name: 锁名
    :param acquire_timeout: 加锁超时时间 单位s
    :param lock_timeout: 锁的原始租期 单位ms
    :return:
    """
    end = time.time() + acquire_timeout if acquire_timeout > 0 else float('inf')

    acquired = False
    identifier: str = identifier
    lock_name = lock_name
    lock_timeout = int(math.ceil(lock_timeout))

    # 判断是否超时监测
    def acquire_lock():
        acquire_lock_lua_result = _acquire_re_entrant_lock_with_timeout_lua(conn, [lock_name], [30 * 1000 if lock_timeout == -1 else lock_timeout, identifier])

        def result(*args, **kwargs):
            return acquire_lock_lua_result

        return result  # 如果不是返回ttl，则成功拿到锁

    # cant write this
    if end < float('inf'):
        acquire_lock = _watch_operation_time(limit_time=int(end - time.time()))(acquire_lock)

    lua_result = acquire_lock()
    if lua_result(default=-1) == -1:
        # 请求锁超时,但是依然将这个future对象传出
        return lua_result
    else:
        if lua_result(default=-1) is None:
            # 获取锁成功
            if lock_timeout == -1:
                watch_thread = _WatchThread(target=_renew_re_entrant_lock, args=[conn, lock_name, identifier, 30 * 1000])
                watch_thread.setDaemon(True)
                watch_thread.start()

            def result(*args, **kwargs):
                return identifier

            return result
        elif acquire_timeout != 0:
            while time.time() < end and not acquired:
                lua_result = _acquire_re_entrant_lock_with_timeout_lua(conn, [lock_name], [lock_timeout, identifier])
                acquired = lua_result is None
                time.sleep(0.001 * (not acquired))

            def result(*args, **kwargs):
                return acquired and identifier

            return result
        else:
            def result(*args, **kwargs):
                return acquired and identifier

            return result


"""释放可重入锁"""
_release_re_entrant_lock_lua = _script_load(
    # KEYS: lock name
    # ARGV ：lock_timeout,id

    # 1. 如果exists 结果为 0 ，**标识无锁**，发送释放锁的消息，返回2，释放成功
    # 2. 如果key存在，但是自己不是持锁者，**无权释放**，返回 nil
    # 3. 如果自己是持锁者，直接-1，然后**判断是否还有重入锁，刷新过期时间**，返回0
    # 4. 如果**无其他重入锁，则删除key**，释放锁，并发送释放消息，并返回1，释放成功
    """
    if (redis.call('exists',KEYS[1]) == 0) then
        return 2
    end
    if (redis.call('hexists',KEYS[1],ARGV[2])==0) then
        return nil
    end
    local counter = redis.call('hincrby',KEYS[1],ARGV[2],-1)
    if (counter > 0) then
        redis.call('pexpire',KEYS[1],ARGV[1])
        return 0
    else
        redis.call('del',KEYS[1])
        return 1
    end
    return nil
    """
)


def release_re_entrant_lock(conn: Redis, lock_name: str, identifier: str, lock_timeout: int = 30 * 1000) -> typing.Union[int, None]:
    """
    释放可重入锁
    :param conn:
    :param lock_name:
    :param identifier:
    :param lock_timeout:
    :return: 0：释放了自己的锁,当还有别的可重入锁 1：释放成功 2:无锁 None:无权释放
    """
    lock_name = lock_name
    return _release_re_entrant_lock_lua(conn, [lock_name], [lock_timeout, identifier])


__all__ = ["acquire_re_entrant_lock_with_timeout", "release_re_entrant_lock", "acquire_lock_with_timeout", "release_lock"]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fast_job', 'fast_job.sonyflake']

package_data = \
{'': ['*']}

install_requires = \
['APScheduler>=3.8.1,<4.0.0', 'fastapi>=0.74.1', 'redis>=4.1.4,<5.0.0']

setup_kwargs = {
    'name': 'fast-job',
    'version': '0.1.11',
    'description': 'a distributed scheduled task scheduling component written for fast-api',
    'long_description': '### fast_job \n- name = "fast_job"\n- description = "Provides scheduling apis and scheduling and task-related services"\n- authors = ["Euraxluo <euraxluo@qq.com>"]\n- license = "The MIT LICENSE"\n- repository = "https://github.com/Euraxluo/fast_job"\n- coverage : 74%\n- version : 0.1.*\n\n![test-report](https://gitee.com/Euraxluo/images/raw/master/pycharm/MIK-HQpicL.png)\n\n#### install\n`pip install fast-job`\n\n#### UseAge\n\n1.wrapper function to build task\n\n```\nfrom fast_job import *\n\n\n@schedule.task(\'task1\', summer="test_task_1", tag=\'test\', description="test_task_1")\ndef test(tag: int):\n    print({"msg": "test_task_1", "tag": tag})\n    return {"msg": "test_task_1", "tag": tag}\n\n\n@schedule.task(\'task2\', summer="test_task_2", tag=\'test\', description="test_task_2")\ndef test2(tag: int):\n    print({"msg": "test_task_2", "tag": tag})\n    return {"msg": "test_task_2", "tag": tag}\n\n\n@schedule.task(\'task3\', summer="test_task_3", tag=\'test\', description="test_task_3")\ndef task3(tag: int):\n    raise Exception(str({"msg": "test_task_2", "tag": tag}))\n```\n\n2.include in your fastApi\n\n```python\nfrom loguru import logger\nfrom fastapi import FastAPI\nfrom example.jobs import schedule, fast_job_api_router\nfrom example.conftest import rdb as redis\n\napp = FastAPI()\n\n\n@app.on_event("startup")\nasync def registry_schedule():\n    schedule.setup(prefix=\'test:\', logger=logger, redis=redis, distributed=True)\n\n\n@app.on_event("shutdown")  # 关闭调度器\nasync def shutdown_connect():\n    schedule.shutdown()\n\n\nprefix = "/test"\napp.include_router(fast_job_api_router, prefix=prefix, tags=["jobs"])  # include router\n```',
    'author': 'Euraxluo',
    'author_email': 'euraxluo@qq.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Euraxluo/fast_job',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
